import { useEffect, useState } from 'react';
import { EMPTY_ARRAY, FilterMatchEnum } from '../constants';
import { isBoolean, trimAndFormatFilterStr } from '../utils';
/**
 * useMenuOptions hook
 *
 * Parse options to array of MenuOptions and perform filtering (if applicable).
 * Set menuOptions state (ensure array returned).
 */
const useMenuOptions = (options, debouncedInputValue, filterMatchFrom, selectedOption, getOptionValue, getOptionLabel, getIsOptionDisabledRef, getFilterOptionStringRef, filterIgnoreCase = false, filterIgnoreAccents = false, isMulti = false, async = false, hideSelectedOptions) => {
    const [menuOptions, setMenuOptions] = useState(EMPTY_ARRAY);
    // Prevent effect from executing on search input mutations in 'async' mode (also prevents filtering from executing)
    const searchValue = !async ? debouncedInputValue : '';
    const hideSelectedOptionsOrDefault = isBoolean(hideSelectedOptions) ? hideSelectedOptions : isMulti;
    useEffect(() => {
        const isFilterMatchAny = filterMatchFrom === FilterMatchEnum.ANY;
        const normalizedSearch = trimAndFormatFilterStr(searchValue, filterIgnoreCase, filterIgnoreAccents);
        const selectedHash = selectedOption.length ? new Set(selectedOption.map((x) => x.value)) : undefined;
        const isOptionFilterMatch = (option) => {
            const optionStr = getFilterOptionStringRef(option);
            const normalizedOptionLabel = trimAndFormatFilterStr(optionStr, filterIgnoreCase, filterIgnoreAccents);
            return isFilterMatchAny
                ? normalizedOptionLabel.indexOf(normalizedSearch) > -1
                : normalizedOptionLabel.substr(0, normalizedSearch.length) === normalizedSearch;
        };
        const parseMenuOption = (data) => {
            const value = getOptionValue(data);
            const label = getOptionLabel(data);
            const menuOption = {
                data,
                value,
                label,
                ...(getIsOptionDisabledRef(data) && { isDisabled: true }),
                ...(selectedHash?.has(value) && { isSelected: true })
            };
            if ((normalizedSearch && !isOptionFilterMatch(menuOption)) ||
                (hideSelectedOptionsOrDefault && menuOption.isSelected)) {
                return;
            }
            return menuOption;
        };
        const nextMenuOptions = options.reduce((acc, option) => {
            const menuOption = parseMenuOption(option);
            menuOption && acc.push(menuOption);
            return acc;
        }, []);
        setMenuOptions(nextMenuOptions);
    }, [
        options,
        searchValue,
        getOptionValue,
        getOptionLabel,
        selectedOption,
        filterMatchFrom,
        filterIgnoreCase,
        filterIgnoreAccents,
        getIsOptionDisabledRef,
        getFilterOptionStringRef,
        hideSelectedOptionsOrDefault
    ]);
    return menuOptions;
};
export default useMenuOptions;
